=head1 NAME

Model::Table - Class for Table Based Statistical Models

=head1 DESCRIPTION

Model::Table is an implementation of the Model class.  The mean value is
computed by a table lookup; if the value is not found in the table, it is
interpolated using simple linear interpolation.  Interpolation will only be
done if the point to be interpolated lies between two known points.  Standard
deviations are interpolated similarly.  Percentiles are computed assuming a
normal distribution.

=head1 APPENDIX

When load() is called on a file, in addition to a "[Table]" line, the file
is expected to have two name/value pairs:

=head2 p_units

    p_units = STRING

p_units is a string containing the name of the predictor variable.

=head2 r_units

    r_units = STRING

r_units is a string containing the name of the response variable.

=head1 TABLE

Furthermore, after these name/value pairs, there should be a table that looks
like

    [Data]
    P1    R1    SD1
    P2    R2    SD2
    P3    R3    SD3
    ...

where the P values are the predictor variable, the R values represent the mean
at that specific predictor and the SD values represent the standard deviation.
If not known, the SD values may be omitted.

=head1 EXAMPLE

The following file predicts means in grams for weeks 12-15 as well as standard
deviations and percentiles for weeks 14-15.

    [Table]
    p_units = weeks
    r_units = g
    [Data]
    12 20
    13 31
    14 49 12
    15 75 17

=cut

package Model::Table;
@ISA = ("Model");

use Statistics::Distributions;
use strict;
use warnings;

sub load {
    my($cn, $fn) = @_;
    my $this = bless {};

    my($fh);
    open($fh, "<$fn") || die "Unable to open $fn: $!\n";

    my $istable = 0;
    my $isdata = 0;
    my $ln = 0;
    my(@pulist, @pslist, @ulist, @slist);
    foreach my $line (<$fh>) {
	chomp($line);
	++$ln;

	$line =~ s/#.*//;

	if($line =~ m/^\s*$/) {
	    next;
	}

	if(! $istable) {
	    if($line =~ m/^\s*\[Table\]\s*$/) {
		$istable = 1;
		next;
	    } else {
		last;
	    }
	}

	if(! $isdata) {
	    if($line =~ m/^\s*\[Data\]\s*$/) {
		$isdata = 1;
		next;
	    }

	    if($line =~ m/^\s*(\S+)\s*=\s*(.*)$/) {
		if($1 eq "p_units") {
		    $this->{"p_units"} = $2;
		} elsif($1 eq "r_units") {
		    $this->{"r_units"} = $2;
		} else {
		    die "$fn:$ln: unknown variable name ($1)\n";
		}
	    } else {
		die "$fn:$ln: unable to parse name/value pair\n";
	    }

	    next;
	}

	if($line =~ m/^\s*(\S+)\s+(\S+)\s+(\S+)\s*$/) {
	    if(! _is_num($1) || ! _is_num($2) || ! _is_num($3)) {
		die "$fn:$ln should contain 2 or 3 numbers\n";
	    }

	    push @pulist, $1;
	    push @pslist, $1;
	    push @ulist, $2;
	    push @slist, $3;
	} elsif($line =~ m/^\s*(\S+)\s+(\S+)\s*$/) {
	    if(! _is_num($1) || ! _is_num($2)) {
		die "$fn:$ln should contain 2 or 3 numbers\n";
	    }

	    push @pulist, $1;
	    push @ulist, $2;
	} else {
	    die "$fn:$ln: line should contain 2 or 3 numbers\n";
	}
    }

    if(! $istable) {
	die "$fn: This does not seem to be a Model::Table " .
	    "file.  There is no '[Table]' declaration.\n";
    }

    if(@pulist) {
	my $last_val = $pulist[0];
	for(my $i = 1; $i < @pulist; ++$i) {
	    if($last_val > $pulist[$i]) {
		die "$fn: data table not sorted in ascending predictor order\n";
	    }
	}

	$this->{"u_min"}  = $pulist[0];
	$this->{"u_max"}  = $pulist[@pulist - 1];
	$this->{"pulist"} = \@pulist;
	$this->{"ulist"}  = \@ulist;
    } else {
	die "$fn: must contain at least one line of data\n";
    }

    if(@pslist) {
	$this->{"s_min"}  = $pslist[0];
	$this->{"s_max"}  = $pslist[@pslist - 1];
	$this->{"pslist"} = \@pslist;
	$this->{"slist"}  = \@slist;
    }

    return $this;
}

sub predict {
    my($this, $x) = @_;

    if($x < $this->{"u_min"} || $x > $this->{"u_max"}) {
	return 0;
    }

    my $ret = _interpolate($this->{"pulist"}, $this->{"ulist"}, $x);

    if($ret) {
	return $ret;
    } else {
	return "0 but true";
    }
}

sub sd {
    my($this, $x) = @_;

    if(!$this->{"pslist"} || $x < $this->{"s_min"} || $x > $this->{"s_max"}) {
	return 0;
    }

    my $ret = _interpolate($this->{"pslist"}, $this->{"slist"}, $x);

    if($ret) {
	return $ret;
    } else {
	return "0 but true";
    }
}

sub percentile {
    my($this, $p, $r) = @_;

    if($p < $this->{"u_min"} || $p  > $this->{"u_max"} ||
       !$this->{"pslist"} ||
       $p < $this->{"s_min"} || $p > $this->{"s_max"}) {
	return 0;
    }

    my $ret = (1 - Statistics::Distributions::uprob(
	    ($r - _interpolate($this->{"pulist"}, $this->{"ulist"}, $p))
	    / _interpolate($this->{"pslist"}, $this->{"slist"}, $p))) * 100;

    if($ret) {
	return $ret;
    } else {
	return "0 but true";
    }
}

sub p_units {
    my($this) = @_;
    return $this->{"p_units"};
}

sub r_units {
    my($this) = @_;
    return $this->{"r_units"};
}

sub _interpolate {
    my($p, $r, $x) = @_;

    my $start = 0;
    my $end = @$p - 1;

    # There is only one response element; special case
    if($start == $end) {
	return $r->[$start];
    }

    while($start + 1 != $end) {
	my $half = $start + int(($end - $start) / 2);
	if($p->[$half] < $x) {
	    $start = $half;
	} else {
	    $end = $half;
	}
    }

    return $r->[$start] + ($x - $p->[$start]) *
		($r->[$end] - $r->[$start]) / ($p->[$end] - $p->[$start]);
}

sub _is_num {
    my($string) = @_;

    if($string =~ m/^[-+]?(?:\d+|\d*\.\d+|\d+\.\d*)(?:[eE][+-]?\d+)?$/) {
	return 1;
    } else {
	return 0;
    }
}

1;
